/***************************************************************
 * game loop and related game code
 *
 * written by Doug Lanford, opus@opusgames.com
 * for more info, go to www.opusgames.com
 *
 * This file is provided as freeware
 **************************************************************/


#include "game.h"
#include "gba_disp.h"
#include "gba_palette.h"
#include "bkg.h"
#include "object.h"
#include "interrupt.h"
#include "math.h"
#include "world.h"
#include "sprite.h"
#include "ship.h"
#include "text.h"
#include "buttons.h"
#include "dat.h"

extern s24f8_vect2D shipSize;
extern s24f8_vect2D playerPos;
extern u16 playerLife;
extern u16 playerShield;
extern u16 numActiveShips;


BackgroundStruct bkg0;

// number of vblanks the previous frame took
u16 frameTime = 0;
u16 frameTimeAccum = 0;	// vblank counter for current frame
s24f8 frameTimeF = 0;

s24f8 screenXf = 0, screenYf = 0;
s16 screenX = 0, screenY = 0;

u16 gameplayState = 0;

// player's score!
u16 killCount = 0;

u8 clockTicks = 0;
u8 seconds = 0;
u8 minutes = 0;
u8 hours = 0;

// game text
u8 msg_pause[] = "-PAUSE-";
u8 msg_playerdead[] = "YOU LOSE";

u8 msg_gamename[] = "STAR-SHOT";
u8 msg_controls_A[] = "A: SHOOT";
u8 msg_controls_B[] = "B: SHIELD";
u8 msg_controls_L[] = "L: STRAFE";
u8 msg_controls_R[] = "R: TURBO";
u8 msg_controls_START[] = "START: PAUSE";
u8 msg_controls_SELECT[] = "SELECT: RESET GAME";

u8 msg_pressStart[] = "PRESS START TO PLAY";


/*************************************************************
 * load VRAM for logo screen
 ************************************************************/
void Game_InitShell (void)
{
u32 *pLoad32, *pCur32;
u16 *pLoad16, *pCur16;
u32 randomSeed = 1;
u32 i, j;

	// set the display for permanent vblank, so we can load VRAM
	REG_DISPLAY_MODE = DISP_MODE_4 | DISP_FORCE_BLANK;

	// load the palette to VRAM
	pLoad32 = (u32 *)VRAM_PALETTES_TOP;
	pCur32 = (u32 *)LogoOpusGames_palette;
	for (i=0; i<128; i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// load the bitmap
	pLoad16 = (u16 *)VRAM_TOP;
	pCur16 = (u16 *)LogoOpusGames_data;
	for (i=0; i<160; i++)
	{
		for (j=0; j<120; j++)
		{
			*pLoad16++ = *pCur16++;
		}
	}

	// set the display for basic tile diplay (mode 0) and enable tile map 0
	REG_DISPLAY_MODE = DISP_MODE_4 | DISP_MAP_2;

	frameTimeAccum = 0;

	// start the interrupts (vblank only)
	Interrupt_Enable(INTERRUPT_VBLANK);

	while(frameTimeAccum < 240);	// wait 4 seconds, then go into game

	// halt all interrupts
	Interrupt_Disable(0);

	frameTimeAccum = 0;

	// load the sprite palette to VRAM
	// load the test sprite palette to VRAM
	pLoad32 = (u32 *)VRAM_SPRITE_PALETTES_TOP;
	pCur32 = (u32 *)ship0_palette;
	for (i=0; i<128; i++)
	{
		*pLoad32++ = *pCur32++;
	}
	pLoad16 = (u16 *)VRAM_SPRITE_PALETTES_TOP;
	pLoad16[254] = 0x001F;
	pLoad16[255] = 0x0000;

	// load sprite tiles (we need the letters for the start screen)
	pLoad32 = (u32 *)VRAM_SPRITE_TILES;

	// load the ship tiles to sprite VRAM (5 ships, each 4 tiles)
	pCur32 = (u32 *)ship0_tiles;
	for (i=0; i<(5*4*16); i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// load the bullet tiles to sprite VRAM (4 bullet frames, each 1 tile)
	pCur32 = (u32 *)bullet0_tiles;
	for (i=0; i<(4*16); i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// load the explosion tiles to sprite VRAM (15 frames, each 4 tile)
	pCur32 = (u32 *)exploA_00_tiles;
	for (i=0; i<(15*4*16); i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// load the number, letter, and icon tiles (10 numbers, 28 letters+, 4 icons)
	pCur32 = (u32 *)nums_tiles;
	for (i=0; i<((10+28+4)*16); i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// init the sprites
	Sprite_InitSystem();

	// set the display
	REG_DISPLAY_MODE = DISP_MODE_0 | DISP_SPRITE_MAPPING_1D | DISP_SPRITES;

	// start the interrupts (vblank only)
	Interrupt_Enable(INTERRUPT_VBLANK);

	// shell game loop
	while ((buttonsDown & BUTTON_START) == 0)
	{
		// make sure at least one vblank occured during last frame
		while (frameTimeAccum == 0)
		{
			// adjust seed value for the random number generator (this ensures more
			// randomness in the generator once we get into gameplay)
			randomSeed++;
		}

		// set the new frame counters
		frameTimeAccum = 0;

		// get button input
		Buttons_Update();

		// reset the sprite allocation for this frame
		Sprite_TopOfFrame();

		Text_DrawString(76, 16, msg_gamename);

		Text_DrawString(16, 48, msg_controls_A);
		Text_DrawString(16, 60, msg_controls_B);
		Text_DrawString(16, 72, msg_controls_L);
		Text_DrawString(16, 84, msg_controls_R);
		Text_DrawString(16, 96, msg_controls_START);
		Text_DrawString(16, 108, msg_controls_SELECT);

		Text_DrawString(48, 144, msg_pressStart);

		// tell system to load the current sprite frame next vblank
		Sprite_EndOfFrame();
	}

	// set the display for permanent vblank, so we can load VRAM
	REG_DISPLAY_MODE = DISP_MODE_4 | DISP_FORCE_BLANK;

	// halt all interrupts
	Interrupt_Disable(0);

	frameTimeAccum = 0;

	// clear VRAM
	pLoad16 = (u16 *)VRAM_TOP;
	for (i=0; i<160; i++)
	{
		for (j=0; j<120; j++)
		{
			*pLoad16++ = 0;
		}
	}

	// init the random number generator
	Random_ReSeed(randomSeed);
}


/*************************************************************
 * load VRAM for gameplay
 ************************************************************/
void Game_LoadVRAMforGameplay (void)
{
u32 *pLoad32, *pCur32;
u16 *pLoad16, *pCur16;
u32 i;

	// set the display for permanent vblank, so we can load VRAM
	REG_DISPLAY_MODE = DISP_MODE_0 | DISP_FORCE_BLANK;

	// load the test map palette to VRAM
	pLoad32 = (u32 *)VRAM_MAP_PALETTES_TOP;
	pCur32 = (u32 *)bkg_palette;
	for (i=0; i<128; i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// load the test sprite palette to VRAM
	pLoad32 = (u32 *)VRAM_SPRITE_PALETTES_TOP;
	pCur32 = (u32 *)ship0_palette;
	for (i=0; i<128; i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// setup background 0 with tile block 0, tile map block 31, 32x32 tiles size
	// this is the background star plane
	Bkg_Init(0, 0, 31, 0, 0, 0, &bkg0);

	// load the stars tiles
	pLoad32 = (u32 *)bkg0.pTiles;
	pCur32 = (u32 *)tileset_Start;
	for (i=0; i<(NUM_TILES*8); i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// load the stars tile map
	pLoad16 = (u16 *)bkg0.pMap;
	pCur16 = (u16 *)stars_layer_00;
	for (i=0; i<(32*32); i++)
	{
		*pLoad16++ = *pCur16++;
	}
}


/*************************************************************
 * init the game setup
 ************************************************************/
void Game_Init (void)
{
	Game_LoadVRAMforGameplay();

	// init the object system
	Object_InitSystem();

	// init the sprites
	Sprite_InitSystem();

	// init the game world
	World_Init(0, 0, int_to_s24f8(1024), int_to_s24f8(1024));

	// init the spaceship settings
	Ship_InitSystem();

	// launch a player object
	{
		s24f8_vect2D tmpPos;

		tmpPos.x = gWorldPos.x + shipSize.x + PLAYER_SCREEN_OFFSET_X;
		tmpPos.y = gWorldPos.y + shipSize.y + PLAYER_SCREEN_OFFSET_Y;

		Ship_New(&tmpPos, SHIP_PLAYER);
	}

	// set the display for basic tile diplay (mode 0) and enable tile map 0
	REG_DISPLAY_MODE = DISP_MODE_0 | DISP_MAP_0 | DISP_SPRITE_MAPPING_1D | DISP_SPRITES;

	// set the tranparency of the backgrounds
	// I will need this when I set up two backgrounds
//	REG_BLEND_SETTINGS = BLEND_FIRST_BG0 | BLEND_SECOND_BG1 | BLEND_SECOND_OBJ | BLEND_ALPHA_NORMAL;
//	SetBlendLevel(8, 8);

	// start the interrupts (vblank only)
	Interrupt_Enable(INTERRUPT_VBLANK);

	// put us in running gameplay mode
	gameplayState = GAMEPLAY_NORMAL;

	frameTimeAccum = 0;

	killCount = 0;

	clockTicks = 0;
	seconds = 0;
	minutes = 0;
	hours = 0;
}


/**************************************************************
 * the master game loop
 *
 * NOTE:	one pass through the loop = one game frame
 *************************************************************/
void Game_MainLoop (void)
{
	// show the shell screen
	Game_InitShell();

	// then go into game
	Game_Init();

	while (1)
	{
		// make sure at least one vblank occured during last frame
		while (frameTimeAccum == 0);

		// set the new frame counters
		frameTime = frameTimeAccum;
		frameTimeF = int_to_s24f8(frameTime);
		frameTimeAccum = 0;

		// adjust game clock
		clockTicks += frameTime;
		if (clockTicks >= 60)
		{
			clockTicks -= 60;
			seconds++;

			if (seconds >= 60)
			{
				seconds -= 60;
				minutes++;

				if (minutes >= 60)
				{
					minutes -= 60;
					hours++;
				}
			}
		}

		// reset the sprite allocation for this frame
		Sprite_TopOfFrame();

		switch (gameplayState)
		{
		case GAMEPLAY_NORMAL:
			// draw the HUD
			Sprite_LoadNextSlot(SPRITE_COLOR_MODE_256 + 8, SPRITE_SIZE_8 + 8, SKULL_TILENUM<<1);
			Text_DrawNumber(16, 8, killCount);

			Sprite_LoadNextSlot(SPRITE_COLOR_MODE_256 + 144, SPRITE_SIZE_8 + 200, HEART_TILENUM<<1);
			Text_DrawNumber(208, 144, playerLife);

			Sprite_LoadNextSlot(SPRITE_COLOR_MODE_256 + 144, SPRITE_SIZE_8 + 8, BOX_TILENUM<<1);
			Text_DrawNumber(16, 144, playerShield);

			Text_DrawNumber(176, 8, hours);
			Text_DrawNumber(196, 8, minutes);
			Text_DrawNumber(216, 8, seconds);
			Text_DrawString(190, 8, ":");
			Text_DrawString(210, 8, ":");
			Sprite_LoadNextSlot(SPRITE_COLOR_MODE_256 + 8, SPRITE_SIZE_8 + 168, CLOCK_TILENUM<<1);

			if (playerLife <= 0)
			{	// display the "you're dead" text
				Text_DrawString(96, 76, msg_playerdead);
			}

			// occasionally create a new ai ship
			if (numActiveShips < 16)
			{	// max of 16 ships active at once
				if ((Random_Get32() % (5*60)) == 0)
				{	// time to launch a new ship
					ANGLE launchAng = Random_Get32() % ANG_360;
					s24f8_vect2D launchPos;

					launchPos.x = playerPos.x + (s24f8_mulInt(COS_TABLE[launchAng], 140));
					launchPos.y = playerPos.y - (s24f8_mulInt(SIN_TABLE[launchAng], 140));

					// randomly pick thet type of ship
					Ship_New(&launchPos, (Random_Get32() % (NUM_SHIP_TYPES-1)) + 1);
				}
			}

			// get joypad input
			Buttons_Update();

			if (buttonsDown & BUTTON_SELECT)
			{	// temp - select button restarts the game from the beginning
				Game_Init();
			}

			if (buttonsDown & BUTTON_START)
			{	// ultra-simple pause mode
				gameplayState = GAMEPLAY_PAUSED;
			}

			// update all the active game objects
			Object_UpdateAll();

			break;

		case GAMEPLAY_PAUSED:
			// just wait for unpause
			Buttons_Update();
			if (buttonsDown & BUTTON_START)
			{
				gameplayState = GAMEPLAY_NORMAL;
			}

			// put up the paused message
			Text_DrawString(96, 76, msg_pause);
			break;

		default:
			break;
		}

		// tell system to load the current sprite frame next vblank
		Sprite_EndOfFrame();
	}
}


