/******************************************************************
 * computer controlled ship game object
 *
 * written by Doug Lanford, opus@opusgames.com
 * for more info, go to www.opusgames.com
 *
 * This file is provided as freeware
 ******************************************************************/

#include "ship.h"
#include "gba_palette.h"
#include "sprite.h"
#include "game.h"
#include "math.h"
#include "world.h"
#include "buttons.h"
#include "bullet.h"
#include "explosion.h"
#include "collide.h"
#include "powerup.h"

extern s16 screenX, screenY;
extern u16 killCount;

s24f8_vect2D playerPos = {0,0};
u16 playerLife = 0;
u16 playerShield = 0;
s24f8 shieldTimer = 0;

s24f8_vect2D shipSize = {int_to_s24f8(16), int_to_s24f8(16)};


ShipSetupStruct shipSetups[NUM_SHIP_TYPES] =
{
	{25, SHIP0_TILENUM, 0x100, PLAYER_ACCEL},
	{5,  SHIP1_TILENUM, 0x100, SHIP1_ACCEL},
	{10, SHIP2_TILENUM, 0x100, SHIP1_ACCEL},
	{5,  SHIP3_TILENUM, 0x100, SHIP1_ACCEL},
};


u16 numActiveShips = 0;

/******************************************************************
 * init the ship setups
 *****************************************************************/
void Ship_InitSystem (void)
{
	numActiveShips = 0;
}


/******************************************************************
 * launch a new ship object
 *
 * IN:		position of the ship
 * 			type of ship
 *
 * RETURNS:	ptr to a new ship object
 *****************************************************************/
ShipStruct *Ship_New (s24f8_vect2D *pPos, u16 type)
{
u32 objID;
ShipStruct *pShip;

	// grab a new object
	objID = Object_New(OBJECT_SHIP);
	if (objID == OBJECT_ID_NULL)
	{	// error... no unused objects left
		return(NULL);
	}

	// get the pointer to it
	pShip = (ShipStruct *)Object_getPtrS(objID);

	// init it
	pShip->shipType = type;

	pShip->pos = *pPos;

	pShip->speed.x = 0;
	pShip->speed.y = 0;
	pShip->speedMag = 0;
	pShip->targetSpeed = int_to_s24f8(1);

	pShip->shipAngle = 0;
	pShip->targetAngle = 0;
	pShip->speedAngle = 0;

	pShip->life = shipSetups[type].maxLife;

	pShip->timer = 0;

	if (type == SHIP_PLAYER)
	{
		playerLife = shipSetups[SHIP_PLAYER].maxLife;
		playerPos = *pPos;
		playerShield = 10;
	}

	pShip->shieldCycle = 0;

	numActiveShips++;

	return(pShip);
}


extern s24f8_vect2D playerPos;

/******************************************************************
 * update a ship object
 *****************************************************************/
u32 Ship_Update (ObjectStruct *pObj)
{
ShipStruct *pShip = (ShipStruct *)pObj;
ShipSetupStruct *pSetup = &shipSetups[pShip->shipType];
u16 drawTile;

	if (pShip->life <= 0)
	{	// dead ship...
		if (pShip->timer == 0)
		{	// setup
			pShip->timer = int_to_s24f8(60);
			pShip->flags = 0;
		}
		else
		{	// make explos until timer runs down
			pShip->timer -= frameTimeF;
			if (pShip->timer <= 0)
			{	// kill the ship
				numActiveShips--;
				killCount++;

				// launch a health powerup
				Powerup_New(&pShip->pos, pSetup->maxLife>>1);

				return(1);
			}

			pShip->flags++;

			// make an explosion
			if (pShip->flags > 10)
			{
				s24f8_vect2D delta;

				pShip->flags = 0;

				delta.x = pShip->pos.x + ((Random_Get32() % int_to_s24f8(16)) - int_to_s24f8(8));
				delta.y = pShip->pos.y + ((Random_Get32() % int_to_s24f8(16)) - int_to_s24f8(8));

				Explosion_New(&delta);
			}
		}

		// set the tile to display with
		drawTile = pSetup->tile;
	}
	else
	{
		if (pShip->shotTimer > 0)
		{	// time down till next time we can shoot
			pShip->shotTimer -= frameTimeF;
		}

		switch(pShip->shipType)
		{
		case SHIP_PLAYER:	// player ship type (get aim and movement from player controls)
			// spin ship to d-pad direction
			if (buttonsHeld & BUTTON_UP)
			{
				if (buttonsHeld & BUTTON_RIGHT)
				{	// up + right = 45 degrees
					pShip->targetAngle = ANG_45;
				}
				else if (buttonsHeld & BUTTON_LEFT)
				{	// up + left = -45
					pShip->targetAngle = ANG_315;
				}
				else
				{	// up = 0
					pShip->targetAngle = ANG_0;
				}
			}
			else if (buttonsHeld & BUTTON_DOWN)
			{
				if (buttonsHeld & BUTTON_RIGHT)
				{	// down + right = 135 degrees
					pShip->targetAngle = ANG_135;
				}
				else if (buttonsHeld & BUTTON_LEFT)
				{	// down + left = -135
					pShip->targetAngle = ANG_225;
				}
				else
				{	// down = 180
					pShip->targetAngle = ANG_180;
				}
			}
			else if (buttonsHeld & BUTTON_RIGHT)
			{	// right = 90
				pShip->targetAngle = ANG_90;
			}
			else if (buttonsHeld & BUTTON_LEFT)
			{	// left = 270
				pShip->targetAngle = ANG_270;
			}

			if ((buttonsHeld & BUTTON_A) && (pShip->shotTimer <= 0))
			{	// fire a bullet
				s24f8_vect2D bulletSpeed;

				pShip->shotTimer = SHOT_TIME;

				// get player's speed
				bulletSpeed.x = s24f8_mul(SIN_TABLE[pShip->speedAngle], pShip->speedMag);
				bulletSpeed.y = -s24f8_mul(COS_TABLE[pShip->speedAngle], pShip->speedMag);

				bulletSpeed.x += s24f8_mulInt(SIN_TABLE[pShip->shipAngle], 2);
				bulletSpeed.y -= s24f8_mulInt(COS_TABLE[pShip->shipAngle], 2);

				Bullet_New(&pShip->pos, &bulletSpeed);
			}

			if (!(buttonsHeld & BUTTON_L))
			{	// not in strafe mode, so fly in direction of ship
				pShip->speedAngle = pShip->shipAngle;

				if (buttonsHeld & (BUTTON_UP | BUTTON_DOWN | BUTTON_RIGHT | BUTTON_LEFT))
				{	// throttle on?
					if (buttonsHeld & BUTTON_R)
					{	// turbo on?
						pShip->targetSpeed = pSetup->maxSpeed<<1;
					}
					else
					{
						pShip->targetSpeed = pSetup->maxSpeed;
					}
				}
				else
				{
					pShip->targetSpeed = 0;
				}
			}

			if ((playerShield > 0) && (buttonsHeld & BUTTON_B))
			{	// shield on?
				// set the tile to display with
				drawTile = SHIP0_SHIELD_TILENUM;

				// do the shield palette cycle
				pShip->timer -= frameTimeF;
				while (pShip->timer < 0)
				{	// cycle to next color setup
					u16 curCyc, i;
					u16 *pLoad16;

					pShip->timer += int_to_s24f8(8);

					pShip->shieldCycle++;
					if (pShip->shieldCycle >= 5)
					{
						pShip->shieldCycle = 0;
					}

					// now set the colors in the sprite palette
					pLoad16 = (u16 *)VRAM_SPRITE_PALETTES_TOP;
					pLoad16 += 48;
					curCyc = pShip->shieldCycle;
					for (i=0; i<5; i++)
					{
						u16 color = curCyc<<3;
						if (color > 31)
						{
							color = 31;
						}

						*pLoad16++ = (color<<5) + 31;

						curCyc++;
						if (curCyc >= 5)
						{
							curCyc = 0;
						}
					}
				}

				// count down shield settings
				shieldTimer -= frameTimeF;
				if (shieldTimer < 0)
				{
					shieldTimer += int_to_s24f8(30);

					playerShield--;
				}
			}
			else
			{	// draw normally
				drawTile = pSetup->tile;
			}

			break;

		case SHIP_CHASER:	// chaser type AI ship
			pShip->targetAngle = Math_GetAngleFromPoints(&pShip->pos, &playerPos);

			// set the speed the ai ship wants to go
			pShip->targetSpeed = pSetup->maxSpeed;

			pShip->speedAngle = pShip->shipAngle;

			// set the tile to display with
			drawTile = pSetup->tile;

			break;

		case SHIP_RANDOMWALK:	// randomwalk type AI ship
			pShip->timer -= frameTimeF;
			if (pShip->timer < 0)
			{	// time to think again
				pShip->timer += (Random_Get32() % int_to_s24f8(90)) + int_to_s24f8(30);

				pShip->targetAngle = Math_GetAngleFromPoints(&pShip->pos, &playerPos);

				// set the speed the ai ship wants to go
				pShip->targetSpeed = pSetup->maxSpeed;
			}

			pShip->speedAngle = pShip->shipAngle;

			// set the tile to display with
			drawTile = pSetup->tile;

			break;

		case SHIP_ZIGZAG:	// randomwalk type AI ship
			pShip->timer -= frameTimeF;
			if (pShip->timer < 0)
			{	// time to think again
				pShip->timer += int_to_s24f8(30);

				pShip->targetAngle = Math_GetAngleFromPoints(&pShip->pos, &playerPos);
				pShip->targetAngle += (Random_Get32() % ANG_90) - ANG_45;

				// set the speed the ai ship wants to go
				pShip->targetSpeed = pSetup->maxSpeed;
			}

			pShip->speedAngle = pShip->shipAngle;

			// set the tile to display with
			drawTile = pSetup->tile;

			break;

		default:
			numActiveShips--;
			return(1);
		}

		if (pShip->shipAngle != pShip->targetAngle)
		{	// need to spin the ship towards the target angle
			s16 dist = pShip->targetAngle - pShip->shipAngle;		// distance to target angle
			u16 turnSpeed = 4 * frameTime;

			if (dist > 0)
			{
				if ((dist < turnSpeed) || (dist > (ANG_360-turnSpeed)))
				{
					pShip->shipAngle = pShip->targetAngle;
				}
				else if (dist > ANG_180)
				{
					pShip->shipAngle -= turnSpeed;
				}
				else
				{
					pShip->shipAngle += turnSpeed;
				}
			}
			else
			{
				if ((dist > -turnSpeed) || (dist < (turnSpeed-ANG_360)))
				{
					pShip->shipAngle = pShip->targetAngle;
				}
				else if (dist < -ANG_180)
				{
					pShip->shipAngle += turnSpeed;
				}
				else
				{
					pShip->shipAngle -= turnSpeed;
				}
			}
		}

		// adjust speed towards target
		if (pShip->speedMag != pShip->targetSpeed)
		{
			s24f8 dist = pShip->targetSpeed - pShip->speedMag;
			s24f8 accel = s24f8_mul(pSetup->accel, frameTimeF);

			if (ABS(dist) <= accel)
			{	// almost to target
				pShip->speedMag = pShip->targetSpeed;
			}
			else
			{
				if (dist > 0)
				{
					pShip->speedMag += accel;
				}
				else
				{
					pShip->speedMag -= accel;
				}
			}
		}

		{
			s24f8 deltaX = s24f8_mul(SIN_TABLE[pShip->speedAngle], pShip->speedMag);
			s24f8 deltaY = -s24f8_mul(COS_TABLE[pShip->speedAngle], pShip->speedMag);

			// adjust screen player's world pos from speed
			pShip->pos.x += deltaX;
			pShip->pos.y += deltaY;

			if (pShip->shipType == SHIP_PLAYER)
			{	// adjust the world position
				gWorldPos.x += deltaX;
				gWorldPos.y += deltaY;
			}
		}
	}


	if (Sprite_CheckOnScreen(&pShip->pos, &shipSize, &gWorldPos))
	{	// ship is on-screen... update it in the sprite OAM table
		u16 rotSlot;

		rotSlot = Sprite_LoadNextRotSlot(COS_TABLE[pShip->shipAngle],
										 SIN_TABLE[pShip->shipAngle],
										 -SIN_TABLE[pShip->shipAngle],
										 COS_TABLE[pShip->shipAngle]);

		Sprite_GetPosAndUpdate(&pShip->pos, &shipSize, &gWorldPos,
							   SPRITE_COLOR_MODE_256 + SPRITE_ROTATION_SCALE + SPRITE_SIZE_DOUBLE,
							   SPRITE_SIZE_16 + (rotSlot<<SPRITE_ROTDATA_SHIFT),
							   drawTile<<1);
	}

	if (pShip->shipType == SHIP_PLAYER)
	{
		u16 ret;

		// temp code for scrolling the screen
		screenX = s24f8_to_int(pShip->pos.x) & 0x0FF;
		screenY = s24f8_to_int(pShip->pos.y) & 0x0FF;
		playerPos = pShip->pos;

		// run player collision
		if (Collide_PointToShip(&pShip->pos))
		{	// we collided with an ai ship
			if (drawTile != SHIP0_SHIELD_TILENUM)
			{	// shield is off, so we can take damage
				pShip->life--;
				if (pShip->life >= 0)
				{
					playerLife = pShip->life;
				}
				else
				{	// we died!
					pShip->timer = 0;
					playerLife = 0;
				}
			}
		}

		ret = Collide_PointToPowerup(&pShip->pos);
		if (ret)
		{	// we picked up a powerup
			pShip->life += ret;
			if (pShip->life > 100)
			{
				pShip->life = 100;
			}

			playerLife = pShip->life;
		}
	}

	return(0);
}



