/*************************************************************
 * code for running sprites
 *
 * written by Doug Lanford, opus@opusgames.com
 * for more info, go to www.opusgames.com
 *
 * This file is provided as freeware
 ************************************************************/

#include "sprite.h"
#include "world.h"

// RAM copy of the sprite OAM table (copies to VRAM each frame) - double buffered
SpriteOAMStruct spriteOAMtable[2][NUM_OAM_SPRITES];

u8 allocSpriteNum[2];
u8 allocSpriteRotNum[2];
u8 spriteCurBuffer = 0;
u8 loadOAMflag = 0;

/***********************************************************
 * init the sprite system
 **********************************************************/
void Sprite_InitSystem (void)
{
u32 i;

	for (i=0; i<NUM_OAM_SPRITES; i++)
	{	// clear out the OAM table
		spriteOAMtable[0][i].attribute0 = 160;		// off-screen position, so not displayed
		spriteOAMtable[0][i].attribute1 = 240;		// off-screen position, so not displayed
		spriteOAMtable[0][i].attribute2 = 0;
		spriteOAMtable[0][i].attribute3 = 0;

		spriteOAMtable[1][i].attribute0 = 160;		// off-screen position, so not displayed
		spriteOAMtable[1][i].attribute1 = 240;		// off-screen position, so not displayed
		spriteOAMtable[1][i].attribute2 = 0;
		spriteOAMtable[1][i].attribute3 = 0;
	}

	spriteCurBuffer = 0;
	loadOAMflag = 0;
}


/***********************************************************
 * get the next hardware sprite slot, and load it with the
 * input settings
 *
 * IN:		attributes0,1,2 of the sprite
 *
 * RETURNS:	number of the sprite slot used (0xFFFF is error)
 **********************************************************/
u16 Sprite_LoadNextSlot (u16 attribute0, u16 attribute1, u16 attribute2)
{
	if (allocSpriteNum[spriteCurBuffer] < 128)
	{
		SpriteOAMStruct *pSlot = &spriteOAMtable[spriteCurBuffer][allocSpriteNum[spriteCurBuffer]];
		u16 out = allocSpriteNum[spriteCurBuffer];

		pSlot->attribute0 = attribute0;
		pSlot->attribute1 = attribute1;
		pSlot->attribute2 = attribute2;

		allocSpriteNum[spriteCurBuffer]++;
		return(out);
	}

	return(0xFFFF);
}


/***********************************************************
 * get the next hardware sprite slot for rotation values, and
 * load it with the input settings
 *
 * IN:		rotation values
 *
 * RETURNS:	number of the slot used (0xFFFF is error)
 **********************************************************/
u16 Sprite_LoadNextRotSlot (s16 xcos, s16 ysin, s16 xnsin, s16 ycos)
{
	if (allocSpriteRotNum[spriteCurBuffer] < 32)
	{
		SpriteOAMRotateStruct *pSlot = &((SpriteOAMRotateStruct *)&spriteOAMtable[spriteCurBuffer])[allocSpriteRotNum[spriteCurBuffer]];
		u16 out = allocSpriteRotNum[spriteCurBuffer];

		pSlot->xcos = xcos;
		pSlot->ysin = ysin;
		pSlot->xnsin = xnsin;
		pSlot->ycos = ycos;

		allocSpriteRotNum[spriteCurBuffer]++;
		return(out);
	}

	return(0xFFFF);
}


/********************************************************************
 * check if a sprite is on-screen
 *
 * IN:		pPos = ptr to the position of the sprite (in s24f8_vect2D)
 * 			pSize = ptr to the sprite size (in s24f8_vect2D)
 * 			pWorldPos = ptr to the screen position (in s24f8_vect2D)
 *
 * RETURNS:	non-zero if the sprite is on-screen
 *******************************************************************/
u32 Sprite_CheckOnScreen (s24f8_vect2D *pPos, s24f8_vect2D *pSize, s24f8_vect2D *pWorldPos)
{
	if ((pPos->x + pSize->x < pWorldPos->x) || (pPos->x - pSize->x > pWorldPos->x + SCREEN_SIZE_X)
		|| (pPos->y + pSize->y < pWorldPos->y) || (pPos->y - pSize->y > pWorldPos->y + SCREEN_SIZE_Y))
	{	// sprite is off-screen
		return(0);
	}

	return(1);
}


/********************************************************************
 * update a sprite in the sprite OAM table, calculating its screen
 * position from world position
 *
 * IN:		pPos = ptr to the position of the sprite (in s24f8_vect2D)
 * 			pSize = ptr to the sprite size (in s24f8_vect2D)
 * 			pWorldPos = ptr to the screen position (in s24f8_vect2D)
 *			attribute0, attribute1, attribute2 - basic attribute data (without sprite position)
 *
 * NOTE:	assumes the sprite is on-screen!
 *******************************************************************/
void Sprite_GetPosAndUpdate (s24f8_vect2D *pPos, s24f8_vect2D *pSize, s24f8_vect2D *pWorldPos, u16 attribute0, u16 attribute1, u16 attribute2)
{
s16 posX, posY;

	// calculate it's screen position
	posX = s24f8_to_int(pPos->x - pSize->x - pWorldPos->x);
	if (posX < 0)
	{
		posX += 512;
	}

	posY = s24f8_to_int(pPos->y - pSize->y - pWorldPos->y);
	if (posY < 0)
	{
		posY += 256;
	}

	attribute0 |= posY;
	attribute1 |= posX;

	Sprite_LoadNextSlot(attribute0, attribute1, attribute2);
}


/********************************************************************
 * check if a sprite is on-screen, and if so then update it in the
 * sprite OAM table
 *
 * IN:		pPos = ptr to the position of the sprite (in s24f8_vect2D)
 * 			pSize = ptr to the sprite size (in s24f8_vect2D)
 * 			pWorldPos = ptr to the screen position (in s24f8_vect2D)
 *			attribute0, attribute1, attribute2 - basic attribute data (without sprite position)
 *
 * RETURNS:	non-zero if the sprite is on-screen
 *******************************************************************/
u32 Sprite_OnScreenUpdate (s24f8_vect2D *pPos, s24f8_vect2D *pSize, s24f8_vect2D *pWorldPos,
						   u16 attribute0, u16 attribute1, u16 attribute2)
{
s16 posX, posY;

	if ((pPos->x + pSize->x < pWorldPos->x) || (pPos->x - pSize->x > pWorldPos->x + SCREEN_SIZE_X)
		|| (pPos->y + pSize->y < pWorldPos->y) || (pPos->y - pSize->y > pWorldPos->y + SCREEN_SIZE_Y))
	{	// sprite is off-screen
		return(0);
	}

	// sprite is on-screen
	// calculate it's screen position
	posX = s24f8_to_int(pPos->x - pSize->x - pWorldPos->x);
	if (posX < 0)
	{
		posX += 512;
	}

	posY = s24f8_to_int(pPos->y - pSize->y - pWorldPos->y);
	if (posY < 0)
	{
		posY += 256;
	}

	attribute0 |= posY;
	attribute1 |= posX;

	Sprite_LoadNextSlot(attribute0, attribute1, attribute2);

	return(1);
}


/***********************************************************
 * sprite code called at top of each frame
 **********************************************************/
void Sprite_TopOfFrame (void)
{
	// flip to other OAM buffer
	spriteCurBuffer = 1 - spriteCurBuffer;

	// reset to no sprites allocated yet this frame
	allocSpriteNum[spriteCurBuffer] = 0;

	// reset to no rotation sets loaded yet this frame
	allocSpriteRotNum[spriteCurBuffer] = 0;
}


/***********************************************************
 * sprite code called at end of each frame
 **********************************************************/
void Sprite_EndOfFrame (void)
{
u32 i;

	// clear out all the unused slots
	for (i=allocSpriteNum[spriteCurBuffer]; i<NUM_OAM_SPRITES; i++)
	{
		spriteOAMtable[spriteCurBuffer][i].attribute0 = 160;
		spriteOAMtable[spriteCurBuffer][i].attribute1 = 240;
	}

	// tell the system to load the OAM memory
	loadOAMflag = spriteCurBuffer + 1;
}



